import Flutter
import UIKit
import MagicSDK
import MagicSDK_Web3
import PromiseKit


enum EthereumConversionError : Error {
    case failedToConvert(String)
}


public class SwiftMagicFlutterPlugin: NSObject, FlutterPlugin {
    
    let etherInWei = pow(Decimal(10), 18)
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "magic", binaryMessenger: registrar.messenger())
        let instance = SwiftMagicFlutterPlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        switch call.method {
        
        case "initializeMagic" :
            self.initializeMagic(call: call, result: result)
            
        case "loginWithMagicLink":
            self.loginWithMagicLink(call: call, result: result)
            
        case "isLoggedIn":
            self.isLoggedIn(result: result)
            
        case "logout":
            self.logout(result: result)
            
        case "updateEmail":
            self.updateEmail(call:call,result: result)
            
        case "sendTransaction":
            self.sendTransaction(call: call, result: result)
            
        case "getAccount":
            self.getAccount(result: result)
            
        case "getUserBalance":
            self.getUserBalance(call: call, result: result)
            
        case "getMetaData":
            self.getMetaData(result: result)
            
        default:
            result(FlutterMethodNotImplemented)
        }
        
        
    }
    
    private func initializeMagic(call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        let args = call.arguments as! Dictionary<String, Any>
        
        if let  publisherKey = args["publisherKey"] as? String,
           let rpcUrl = args["rpcUrl"] as? String,
           let chainId = args["chainId"] as? String {
            
            let config = CustomNodeConfiguration(
                rpcUrl: rpcUrl,
                chainId: Int(chainId))
            
            Magic.shared = Magic(apiKey: publisherKey, customNode: config)
            result(true)
            
        } else if let  publisherKey = args["publisherKey"] as? String {
            Magic.shared = Magic(apiKey: publisherKey);
            result(true)
        } else {
            result(FlutterError(code: "INVALID_CALL",
                                message: "Parameters unavailable",
                                details: nil))
        }
        
    }
    
    
    private func loginWithMagicLink(call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        if let email = call.arguments as? String {
            
            let config = LoginWithMagicLinkConfiguration(email: email)
            
            Magic.shared.auth.loginWithMagicLink(config).done({ res in
                result(res) // DIDToken
            }).catch({ error in
                result(FlutterError(code: "ERROR",
                                    message: error.localizedDescription,
                                    details: nil))// handle Error
            })
        } else {
            result(FlutterError(code: "INVALID_CALL",
                                message: "Parameters unavailable",
                                details: nil))
        }
        
    }
    
    private func isLoggedIn(result: @escaping FlutterResult) {
        
        Magic.shared.user.isLoggedIn(response: { response in
            guard let magicResult = response.result
            else {
                return result(false as Bool)
            }
            result(Bool(magicResult))
        })
    }
    
    private func getAccount(result: @escaping FlutterResult) {
        let web3 = Web3(provider: Magic.shared.rpcProvider)
        
        firstly {
            // Get user's Ethereum public address
            web3.eth.accounts()
        }.done { accounts -> Void in
            if let account = accounts.first {
                // Set to UILa
                result(try String(account.hex(eip55: false)))
                
            } else {
                print("No Account Found")
            }
        }.catch { error in
            result(FlutterError(code: "ERROR",
                                message: "Error loading accounts and balance: \(error)",
                                details: nil))
        }
    }
    
    private func sendTransaction(call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        
        do {
            
            let args = call.arguments as! Dictionary<String, Any>
            
            let amount = args["amount"] as! String
            //          let gasPrice = args["gasPrice"] as! String
            //          let gasLimit = args["gasLimit"] as! String
            let from = args["from"] as! String
            let to = args["to"] as! String
            
            let toAddress =  EthereumAddress(hexString: to)
            
            let web3 = Web3(provider: Magic.shared.rpcProvider)
            
            let transaction =  EthereumTransaction(
                from: EthereumAddress(hexString: from), // from Get User Info section
                to: toAddress,
                value: EthereumQuantity(quantity: try self.toWei(ether: amount)))
            
            firstly {
                web3.eth.sendTransaction(transaction: transaction)
            }.done { transactionHash in
                result(try String(transactionHash.hex()))
            }.catch { error in
                result(FlutterError(code: "ERROR",
                                    message: "Error making transaction: \(error)",
                                    details: nil))
            }
            
            
            
            
        } catch let error {
            result(FlutterError(code: "ERROR",
                                message: error.localizedDescription,
                                details: nil))
        }
    }
    
    private func getMetaData(result: @escaping FlutterResult) {
        // Assuming user is logged in
        Magic.shared.user.getMetadata(response: { response in
            guard
                let metadata = response.result
            else {
                result(FlutterError(code: "ERROR",
                                    message:response.error!.localizedDescription,
                                    details: nil))
                return
            }
            
            let meta  : [String:Any]  = [
                "issuer": metadata.issuer!,
                "publicAddress": metadata.publicAddress!,
                "email": metadata.email!
            ]
            
            result(meta as Dictionary<String, Any>)
        })
    }
    
    private func updateEmail(call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        if let newEmail = call.arguments as? String {
            // Assuming user is logged in
            let configuration = UpdateEmailConfiguration(email: newEmail)
            
            Magic.shared.user.updateEmail(configuration, response: { response in
                guard
                    let magicResult = response.result
                else {
                    return result(FlutterError(code: "ERROR",
                                               message:response.error!.localizedDescription,
                                               details: nil))
                }
                
                result(Bool(magicResult))
                
            })
        } else {
            result(FlutterError(code: "INVALID_CALL",
                                message: "Parameters unavailable",
                                details: nil))
        }
        
        
    }
    
    
    private func logout(result: @escaping FlutterResult) {
        // Assuming user is logged in
        Magic.shared.user.logout(response: { response in
            guard
                let magicResult = response.result
            else {
                return result(FlutterError(code: "ERROR",
                                           message:response.error!.localizedDescription,
                                           details: nil))
            }
            result(Bool(magicResult))
        })
    }
    
    private func getUserBalance(call: FlutterMethodCall, result: @escaping FlutterResult) {
        
        if let address = call.arguments as? String {
            
            let web3 = Web3(provider: Magic.shared.rpcProvider)
            
            let e = try? EthereumAddress(hex:address, eip55: false)
            
            guard let ethereumAddress = e else {
                return
            }
            
            firstly {
                web3.eth.getBalance(address: ethereumAddress, block: .latest)
                
            }.done { ethereumQuantity in
                
                let eth:Decimal = self.toEther(wei: ethereumQuantity.quantity)
                
                result(eth.description)
            }.catch { error in
                result(FlutterError(code: "ERROR",
                                    message: error.localizedDescription,
                                    details: nil))
            }
            
            
        }
        
    }
    
    
    
    private func toEther(wei: BigUInt) -> Decimal {
        
        
        guard let decimalWei = Decimal(string: wei.description) else {
            return 0
        }
        
        return decimalWei / etherInWei
        
    }
    
    
    private func toWei(ether: Decimal) throws -> BigUInt {
        guard let wei = BigUInt((ether * etherInWei).description) else {
            throw EthereumConversionError.failedToConvert("Failed to convert eth -> wei")
        }
        return wei
    }
    
    private func toWei(ether: String) throws ->BigUInt {
        guard let decimalEther = Decimal(string: ether) else {
            throw EthereumConversionError.failedToConvert("Failed to convert eth -> wei")
        }
        return try toWei(ether: decimalEther)
    }
    
    
}
